# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from PySide6.QtCore import (
    Qt,
    QPointF,
)
from PySide6.QtGui import QFontMetrics
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGraphicsRectItem,
    QGraphicsItem,
    QLineEdit,
    QMenu,
    QMessageBox,
)

from .constants import (
    TIMELINE_TITLE_HEIGHT,
    TIMELINE_TITLE_ON_BG_COLOR,
    TIMELINE_TITLE_OFF_BG_COLOR,
    TIMELINE_TITLE_ON_FG_COLOR,
    TIMELINE_TITLE_OFF_FG_COLOR,
)
from .event import (
    EventCollection,
    ChangeEvent,
    ChooseEvent,
)
from .textedit import TextEdit


class Timeline(QGraphicsRectItem):

    def __init__(self, name: str, time_pane):
        super().__init__()
        self.name = name
        self.time_pane = time_pane
        self.occurrences = []
        self.event_collection = EventCollection()
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)
        self.setAcceptHoverEvents(True)
        self.title = TimelineTitle(self.name, self)
        self._select = False
        self.description = None

    @property
    def select(self):
        return self._select

    @select.setter
    def select(self, select):
        if select != self._select:
            self._select = select

    def add_occurrence(self, occurrence):
        """Add an occurrence to the timeline"""
        self.occurrences.append(occurrence)
        self.occurrences.sort(key=lambda x: x.begin_time)

    def remove_occurrence(self, occurrence):
        """Remove an occurrence from the timeline"""
        if occurrence in self.occurrences:
            self.occurrences.remove(occurrence)
            self.time_pane.scene.removeItem(occurrence)
            self.time_pane.data_needs_save = True

    def update_rect_width(self, new_width: float):
        """Update the width of the timeline"""
        rect = self.rect()
        rect.setWidth(new_width)
        self.setRect(rect)
        self.title.update_rect_width(new_width)

    def on_remove(self):
        if self.occurrences:
            answer = QMessageBox.question(
                self.time_pane,
                "Confirmation",
                "There are occurrences present. "
                "Do you want to remove this timeline?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if answer == QMessageBox.StandardButton.Yes:
                while self.occurrences:
                    self.occurrences[0].remove()
        # The following does not yet work, since there is no provision for
        # adjusting the positions of the timelines inside the time pane.
        # self.time_pane.scene.removeItem(self)
        # if self in self.time_pane.timelines:
        #     self.time_pane.timelines.remove(self)
        # del self

    def edit_properties(self):
        dialog = QDialog(self.time_pane)
        dialog.setWindowTitle("Timeline properties")

        layout = QFormLayout(self.time_pane)
        self.name_edit = QLineEdit()
        self.name_edit.setText(self.title.text)
        self.name_edit.returnPressed.connect(dialog.accept)
        layout.addRow("Name: ", self.name_edit)
        layout.addRow("Description:", None)
        self.description_edit = TextEdit(self, self.description)
        layout.addRow(self.description_edit)
        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Cancel,
            self.time_pane,
        )
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)
        layout.addWidget(buttons)

        dialog.setLayout(layout)

        dialog.exec()
        if dialog.result() == dialog.DialogCode.Accepted:
            is_changed = False
            name = self.name_edit.text()
            if self.title.text != name:
                self.title.text = name
                is_changed = True
            description = self.description_edit.toPlainText()
            if self.description != description:
                self.description = description
                self.title.setToolTip(self.description)
                is_changed = True
            if is_changed:
                self.time_pane.data_needs_save = True

    def update_occurrences(self):
        for occurrence in self.occurrences:
            occurrence.update_style()

    def edit_events(self):
        while True:
            events_dialog = ChooseEvent(self.event_collection, self.time_pane)
            events_dialog.exec()
            if events_dialog.result() == QMessageBox.DialogCode.Accepted:
                e = events_dialog.get_chosen()
                name = e.name
                color = e.color
                description = e.description
                ChangeEvent(e, self.time_pane).exec()
                if (
                    name != e.name
                    or description != e.description
                    or color != e.color
                ):
                    self.update_occurrences()
                    self.time_pane.data_needs_save = True
            if events_dialog.result() == QMessageBox.DialogCode.Rejected:
                break
        if self.event_collection.is_changed():
            self.time_pane.data_needs_save = True

    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.RightButton:
            menu = QMenu()
            menu.addAction("Add new timeline").triggered.connect(
                self.time_pane.handle_timeline
            )
            menu.addAction(
                "Delete timeline (not yet fully implemented)"
            ).triggered.connect(self.on_remove)
            menu.addAction("Edit timeline properties").triggered.connect(
                self.edit_properties
            )
            menu.addAction("Edit events").triggered.connect(self.edit_events)
            menu.exec(event.screenPos())
        else:
            super().mousePressEvent(event)
        return


class TimelineTitle(QGraphicsRectItem):

    def __init__(self, text: str, parent: Timeline = None):
        super().__init__(parent)
        self.text = text
        rect = self.parentItem().rect()
        rect.setHeight(TIMELINE_TITLE_HEIGHT)
        self.setRect(rect)
        self.parent = parent

    def paint(self, painter, option, widget=...):
        # Draw the rectangle
        self.draw_rect(painter)

        # Draw the text
        self.draw_text(painter)

    def draw_rect(self, painter):
        """Draw the timeline title rectangle"""
        # Set Pen and Brush for rectangle
        if self.parent.select:
            color = TIMELINE_TITLE_ON_BG_COLOR
        else:
            color = TIMELINE_TITLE_OFF_BG_COLOR
        painter.setPen(color)
        painter.setBrush(color)
        painter.drawRect(self.rect())

    def draw_text(self, painter):
        """Draw the timeline title text"""
        if self.parent.select:
            color = TIMELINE_TITLE_ON_FG_COLOR
        else:
            color = TIMELINE_TITLE_OFF_FG_COLOR
        painter.setPen(color)
        painter.setBrush(color)

        font = painter.font()
        fm = QFontMetrics(font)

        text_width = fm.boundingRect(self.text).width()
        text_height = fm.boundingRect(self.text).height()
        text_descent = fm.descent()

        # Get timeline polygon based on the viewport
        timeline_in_viewport_pos = self.parentItem().time_pane.view.mapToScene(
            self.rect().toRect()
        )

        bounding_rect = timeline_in_viewport_pos.boundingRect()

        # Get the viewport rect
        viewport_rect = self.parentItem().time_pane.view.viewport().rect()

        # Compute the x position for the text
        x_alignCenter = bounding_rect.x() + viewport_rect.width() / 2

        # No idea why the "-2", in the vertical position, is needed here
        text_position = QPointF(
            x_alignCenter - text_width / 2,
            TIMELINE_TITLE_HEIGHT / 2 + text_height / 2 - text_descent - 2
        )

        painter.drawText(text_position, self.text)

    def update_rect_width(self, new_width):
        rect = self.rect()
        rect.setWidth(new_width)
        self.setRect(rect)



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QRadioButton,
    QLabel,
    QLineEdit,
    QVBoxLayout,
)
from functools import partial
from datetime import datetime


class Coders:
    """Store and manipulate the list of coders"""

    def __init__(self, info, widget=None):
        self.coders = [Coder(item) for item in info]
        self.current = None
        self.dialog = None
        self.widget = widget

    def set_current(self):

        while len(self.coders) == 0:
            self.new_coder()

        while True:

            self.changed = False

            self.dialog = QDialog(self.widget)
            self.dialog.setWindowTitle("Coders")

            layout = QVBoxLayout()
            self.dialog.setLayout(layout)

            label = QLabel("Select a coder identity:")
            layout.addWidget(label)

            for coder in self.coders:
                button = QRadioButton(coder.get_string())
                button.clicked.connect(partial(self.update, coder))
                if self.current == coder:
                    button.setChecked(True)
                layout.addWidget(button)

            buttonbox = QDialogButtonBox(
                QDialogButtonBox.StandardButton.Open
                | QDialogButtonBox.StandardButton.Apply
                | QDialogButtonBox.StandardButton.Ok,
                self.dialog,
            )
            buttonbox.accepted.connect(self.dialog.accept)
            buttonbox.button(QDialogButtonBox.StandardButton.Apply).setText(
                "Edit"
            )
            buttonbox.button(
                QDialogButtonBox.StandardButton.Apply
            ).clicked.connect(self.edit_coder)
            buttonbox.button(QDialogButtonBox.StandardButton.Open).setText(
                "New"
            )
            buttonbox.button(
                QDialogButtonBox.StandardButton.Open
            ).clicked.connect(self.new_coder)
            layout.addWidget(buttonbox)

            self.dialog.exec()
            if self.changed:
                continue
            if self.current:
                break

    def update(self, coder):
        self.current = coder

    def to_list(self):
        return [x.get_dict() for x in self.coders]

    def edit_coder(self):
        if self.current:
            self.current.edit()
        self.changed = True
        self.dialog.accept()

    def new_coder(self):
        coder = Coder({"name": "", "email": ""}, self.widget)
        coder.edit()
        if coder.name != "":
            self.coders.append(coder)
        self.changed = True
        self.current = coder
        if self.dialog:
            self.dialog.accept()


class Coder:
    """Store and manipulate information for individual coder"""

    def __init__(self, info, widget=None):
        self.name = info["name"]
        if "email" in info:
            self.email = info["email"]
        else:
            self.email = None
        if "last-time" in info:
            self.last_time = info["last-time"]
        else:
            self.last_time = None
        self.button = None
        self.widget = widget

    def get_string(self):
        retval = self.name
        if self.email:
            retval += f" <{self.email}>"
        if self.last_time:
            retval += f" ({self.last_time})"
        return retval

    def get_dict(self):
        retval = {"name": self.name}
        if self.email:
            retval["email"] = self.email
        if self.last_time:
            retval["last-time"] = self.last_time
        return retval

    def set_date_now(self):
        self.last_time = datetime.today().strftime("%Y-%m-%d %H:%M:%S")

    def edit(self):

        self.dialog = QDialog(self.widget)
        self.dialog.setWindowTitle("Edit coder information")

        layout = QFormLayout()
        self.dialog.setLayout(layout)

        self.name_edit = QLineEdit()
        self.name_edit.setText(self.name)
        layout.addRow("Name :", self.name_edit)

        self.email_edit = QLineEdit()
        if self.email:
            self.email_edit.setText(self.email)
        self.email_edit.setFixedWidth(300)
        layout.addRow("Email :", self.email_edit)

        buttonbox = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok,
            self.dialog,
        )
        buttonbox.accepted.connect(self.accept)
        layout.addWidget(buttonbox)

        self.dialog.exec()

    def accept(self):
        self.name = self.name_edit.text()
        new_email = self.email_edit.text()
        if new_email != "":
            self.email = new_email
        self.dialog.accept()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from PySide6.QtWidgets import (
    QColorDialog,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QHBoxLayout,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QWidget,
)
from PySide6.QtGui import (
    QColor,
    QIcon,
)
from PySide6.QtWidgets import QStyle
from functools import partial

from .assets import Assets
from .constants import EVENT_DEFAULT_COLOR
from .textedit import TextEdit
from .utils import color_fg_from_bg


class Event:

    def __init__(self, name=None, color=None, description=None):
        self._name = name
        self._color = color
        self._description = description

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        self._name = new_name

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, new_color):
        self._color = new_color

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, new_description):
        self._description = new_description


class EventCollection:
    def __init__(self):
        self._collection = list()
        self._changed = False

    def __iter__(self):
        self.names = sorted(
            [x.name for x in self._collection], key=str.casefold
        )
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self.names):
            result = self.names[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def set_changed(self):
        self._changed = True

    def is_changed(self):
        return self._changed

    def get_event(self, name):
        for event in self._collection:
            if event.name == name:
                return event
        return None

    def add_event(self, event):
        if not isinstance(event.name, str):
            raise EventCollectionError("The name of an event must be a string")
        if event.name == "":
            raise EventCollectionError("The name of an event must not be empty")
        if event.name in list(iter(self)):
            raise EventCollectionError(
                f"An event with name '{event.name}' already exists"
            )
        else:
            if not isinstance(event.color, QColor):
                raise EventCollectionError("Color must be a QColor object")
            if not isinstance(event.description, str):
                raise EventCollectionError(
                    "The description of an event must be a string"
                )
            self._collection.append(event)


class EventCollectionError(Exception):
    def __init__(self, message):
        super().__init__(message)
        QMessageBox.warning(None, "Error", f"{message}")


class ChooseEvent(QDialog):

    def __init__(self, event_collection, widget=None):
        super().__init__(widget)
        self.chosen = None
        self.event_collection = event_collection
        self.widget = widget
        self.layout = None
        self.create_ui()

    def create_ui(self):
        self.setWindowTitle("Events")
        self.setStyleSheet(
            "QPushButton {"
            "    border: 1px solid gray;"
            "    border-radius: 5px;"
            "    padding: 5px"
            "}"
            "QPushButton:default {"
            "    border: 3px solid black;"
            "}"
        )
        self.re_layout()

        if len(list(iter(self.event_collection))) > 0:
            event_buttons = QHBoxLayout()
            self.buttons = []
            for name in iter(self.event_collection):
                event = self.event_collection.get_event(name)
                button = QPushButton(event.name)
                button.setDefault(len(self.buttons) == 0)
                self.buttons.append(button)
                bg_color = event.color
                fg_color = color_fg_from_bg(bg_color)
                button.setStyleSheet(
                    "QPushButton {"
                    "    background-color: qlineargradient("
                    "        x1:0, y1:0, x2:0, y2:1,"
                    f"       stop:0 {bg_color.name()},"
                    f"       stop:1 {bg_color.name()}"
                    "    );"
                    f"   color: {fg_color.name()};"
                    "}"
                )
                button.setAutoFillBackground(False)
                button.clicked.connect(partial(self.set_chosen, event))
                button.setToolTip(event.description)
                event_buttons.addWidget(button)
            self.layout.addRow(event_buttons)
        control_buttons = QHBoxLayout()
        control_buttons.addStretch()
        assets = Assets()
        new_button = QPushButton(
            QIcon(assets.get("plus.png")),
            "New",
            self,
        )
        new_button.setDefault(False)
        new_button.clicked.connect(self.new_event)
        control_buttons.addWidget(new_button)
        finish_button = QPushButton(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogOkButton),
            "Finish",
            self,
        )
        finish_button.setDefault(False)
        finish_button.clicked.connect(self.reject)
        control_buttons.addWidget(finish_button)
        control_buttons.addStretch()
        self.layout.addRow(control_buttons)

    def re_layout(self):
        if self.layout:
            QWidget().setLayout(self.layout)
        self.layout = QFormLayout(self)

    def set_chosen(self, val):
        self.chosen = val
        self.accept()

    def get_chosen(self):
        return self.chosen

    def new_event(self):
        event = Event("", EVENT_DEFAULT_COLOR, "")
        dialog = ChangeEvent(event, self.widget)
        dialog.exec()
        if dialog.result() == QMessageBox.DialogCode.Accepted:
            if event.name == "":
                QMessageBox.warning(
                    self, "Warning", "Event name cannot be empty"
                )
            else:
                self.event_collection.add_event(event)
                self.event_collection.set_changed()
        self.create_ui()


class ChangeEvent(QDialog):
    def __init__(self, event, widget=None):
        super().__init__(widget)
        self.setWindowTitle("Change event")
        self.event = event
        self.create_ui()

    def create_ui(self):
        layout = QFormLayout(self)
        widgetbox = QHBoxLayout()
        self.name_edit = QLineEdit(self)
        self.name_edit.setText(self.event.name)
        widgetbox.addWidget(self.name_edit)
        self.color_button = QPushButton("color")
        self.color_button.clicked.connect(self.choose_color)
        self.set_style()
        widgetbox.addWidget(self.color_button)
        layout.addRow("Name: ", widgetbox)
        layout.addRow("Description:", None)
        self.description_edit = TextEdit(self, self.event.description)
        layout.addRow(self.description_edit)
        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Cancel,
            self,
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    def accept(self):
        self.event.name = self.name_edit.text()
        self.event.description = self.description_edit.toPlainText()
        super().accept()

    def choose_color(self):
        dialog = QColorDialog(self.event.color, self)
        dialog.exec()
        if dialog.result() == dialog.DialogCode.Accepted:
            self.event.color = dialog.currentColor()
            self.set_style()

    def set_style(self):
        bg_color = self.event.color
        fg_color = color_fg_from_bg(bg_color)
        self.color_button.setStyleSheet(
            "QPushButton {"
            "background-color: qlineargradient(x1:0, y1:0, x2:0, y2:1,"
            f"   stop:0 {bg_color.name()}, stop:1 {bg_color.name()});"
            f" color: {fg_color.name()};"
            "border: 2px solid black;"
            "border-radius: 5px;"
            "padding: 6px"
            "}"
            "QPushButton:hover {"
            "    border: 3px solid black;"
            "}"
        )



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod

from PySide6.QtCore import (
    Qt,
    QRectF,
)
from PySide6.QtGui import (
    QPen,
)
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGraphicsRectItem,
    QGraphicsItem,
    QMenu,
    QMessageBox,
)

from .constants import (
    OCCURRENCE_PEN_COLOR,
    OCCURRENCE_BG_COLOR,
    OCCURRENCE_PEN_WIDTH_ON_CURSOR,
    OCCURRENCE_PEN_WIDTH_OFF_CURSOR,
    OCCURRENCE_HANDLE_WIDTH,
    OCCURRENCE_HANDLE_HEIGHT_FRACTION,
    TIMELINE_HEIGHT,
    TIMELINE_TITLE_HEIGHT,
)
from .event import ChooseEvent
from .textedit import TextEdit
from .utils import color_fg_from_bg


class Occurrence(QGraphicsRectItem):

    def __init__(
        self,
        timeline,
        begin_time: int,
        end_time: int,
        # FIXME: [CREATION] are those really needed?
        lower_bound: int = None,
        upper_bound: int = None,
    ):
        """Initializes the Occurrence widget"""
        super().__init__(timeline)
        self.brush_color = OCCURRENCE_BG_COLOR
        self.pen_color = OCCURRENCE_PEN_COLOR
        self.pen_width = OCCURRENCE_PEN_WIDTH_OFF_CURSOR
        self.event = None
        self.name = None
        self.timeline = timeline
        self.time_pane = timeline.time_pane
        self.mfps = self.time_pane.video.mfps
        self.begin_time = begin_time
        self.end_time = end_time
        factor = self.time_pane.scene.width() / self.time_pane.duration
        begin_x_position = int(self.begin_time * factor)
        end_x_position = int(self.end_time * factor)
        self.selected = False
        self.begin_handle: OccurrenceHandle = None
        self.end_handle: OccurrenceHandle = None
        self.setRect(
            QRectF(
                0,
                0,
                end_x_position - begin_x_position,
                TIMELINE_HEIGHT - TIMELINE_TITLE_HEIGHT,
            )
        )

        self.setX(begin_x_position)
        self.setY(TIMELINE_TITLE_HEIGHT)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.comment: str = ""

    # FIXME: [CREATION] move into Timeline
    def can_be_initiated(occurrences, time):
        """Check if the occurrence can be initiated"""
        lower_bound = upper_bound = None
        valid = True
        occurrence_under_cursor = None

        # Loop through the occurrences of the selected timeline
        for a in occurrences:
            if a.begin_time <= time <= a.end_time:
                valid = False
                occurrence_under_cursor = a
                break
            if not lower_bound:
                if a.end_time < time:
                    lower_bound = a.end_time + int(a.mfps / 2)
            else:
                if a.end_time < time:
                    if lower_bound < a.end_time:
                        lower_bound = a.end_time + int(a.mfps / 2)
            if not upper_bound:
                if a.begin_time > time:
                    upper_bound = a.begin_time - int(a.mfps / 2)
            else:
                if a.begin_time > time:
                    if upper_bound > a.begin_time:
                        upper_bound = a.begin_time - int(a.mfps / 2)
        return valid, lower_bound, upper_bound, occurrence_under_cursor

    def mousePressEvent(self, event):
        if not self.time_pane.occurrence_in_creation:
            self.time_pane.select_timeline(self.timeline)

    def mouseReleaseEvent(self, event):
        return

    def mouseDoubleClickEvent(self, event):
        if not self.time_pane.occurrence_in_creation:
            self.setSelected(True)
            self.get_bounds()
            self.time_pane.select_timeline(self.timeline)

    def focusOutEvent(self, event):
        self.setSelected(False)
        super().focusOutEvent(event)

    def contextMenuEvent(self, event):
        if not self.isSelected():
            super().contextMenuEvent(event)
            return
        can_merge_previous = False
        for occurrence in self.timeline.occurrences:
            if (
                occurrence.end_time == self.begin_time
                and self.name == occurrence.name
            ):
                can_merge_previous = True
                break
        can_merge_next = False
        for occurrence in self.timeline.occurrences:
            if (
                self.end_time == occurrence.begin_time
                and self.name == occurrence.name
            ):
                can_merge_next = True
                break
        menu = QMenu()
        menu.addAction("Delete occurrence").triggered.connect(self.on_remove)
        menu.addAction("Change occurrence event").triggered.connect(
            self.change_event
        )
        if can_merge_previous:
            menu.addAction("Merge with previous occurrence").triggered.connect(
                self.merge_previous
            )
        if can_merge_next:
            menu.addAction("Merge with next occurrence").triggered.connect(
                self.merge_next
            )
        menu.addAction("Comment occurrence").triggered.connect(
            self.edit_comment
        )
        menu.exec(event.screenPos())

    def on_remove(self):
        answer = QMessageBox.question(
            self.time_pane,
            "Confirmation",
            "Do you want to remove the occurrence?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if answer == QMessageBox.StandardButton.Yes:
            self.remove()

    def edit_comment(self):
        comment_dialog = OccurrenceComment(self.comment, self.time_pane)
        comment_dialog.exec()
        if comment_dialog.result() == QMessageBox.DialogCode.Accepted:
            comment = comment_dialog.get_text()
            if self.comment != comment:
                self.comment = comment
                self.setToolTip(self.comment)
                self.time_pane.data_needs_save = True

    def merge_previous(self):
        for occurrence in self.timeline.occurrences:
            if (
                self.begin_time == occurrence.end_time
                and self.name == occurrence.name
            ):
                break
        self.begin_time = occurrence.begin_time
        occurrence.remove()
        self.update_rect()
        self.update()

    def merge_next(self):
        for occurrence in self.timeline.occurrences:
            if (
                self.end_time == occurrence.begin_time
                and self.name == occurrence.name
            ):
                break
        self.end_time = occurrence.end_time
        occurrence.remove()
        self.update_rect()
        self.update()

    def change_event(self):
        events_dialog = ChooseEvent(
            self.timeline.event_collection, self.timeline.time_pane
        )
        events_dialog.exec()
        if events_dialog.result() == QMessageBox.DialogCode.Accepted:
            event = events_dialog.get_chosen()
            if event != self.event:
                self.set_event(event)
                self.update()
                self.time_pane.data_needs_save = True

    def remove(self):
        self.timeline.remove_occurrence(self)
        del self

    def paint(self, painter, option, widget=None):
        # Draw the occurrence rectangle
        self.draw_rect(painter)

        # Draw the name of the occurrence in the occurrence rectangle
        self.draw_name(painter)

        if self.isSelected():
            self.show_handles()
        else:
            self.hide_handles()

    def draw_rect(self, painter):
        """Draw the occurrence rectangle"""
        pen: QPen = QPen(self.pen_color)
        pen.setWidth(self.pen_width)

        if self.isSelected():
            # Set border dotline if occurrence is selected
            pen.setStyle(Qt.PenStyle.DotLine)
        painter.setPen(pen)
        painter.setBrush(self.brush_color)

        # Draw the rectangle
        painter.drawRect(self.rect())

    def draw_name(self, painter):
        """Draws the name of the occurrence"""
        if self.name:
            col = color_fg_from_bg(self.brush_color)
            painter.setPen(col)
            painter.setBrush(col)
            painter.drawText(
                self.rect(), Qt.AlignmentFlag.AlignCenter, self.name
            )

    def set_event(self, event=None):
        """Updates the event"""
        if event is None:
            self.event = None
            self.brush_color = OCCURRENCE_BG_COLOR
        else:
            self.event = event
            self.brush_color = event.color
            self.name = event.name
            self.setToolTip(
                self.comment if self.comment != "" else "(no comment)"
            )
            if self.begin_handle:
                self.begin_handle.setBrush(event.color)
                self.end_handle.setBrush(event.color)

    def update_style(self):
        if self.event:
            self.brush_color = self.event.color
            self.name = self.event.name

    def update_rect(self):
        new_rect = self.time_pane.scene.sceneRect()
        # Calculate position to determine width
        factor = new_rect.width() / self.time_pane.duration
        begin_x_position = self.begin_time * factor
        end_x_position = self.end_time * factor
        self.setX(begin_x_position)

        # Update the rectangle
        rect = self.rect()
        rect.setWidth(end_x_position - begin_x_position)
        self.setRect(rect)

        if self.begin_handle:
            self.begin_handle.setX(self.rect().x())
            self.end_handle.setX(self.rect().width())

    def update_begin_time(self, begin_time: int):
        self.begin_time = begin_time
        self.update_rect()
        self.update()

    def update_end_time(self, end_time: int):
        """Updates the end time"""
        self.end_time = end_time
        self.update_rect()
        self.update()

    def update_selectable_flags(self):
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)
        self.update()

    def create_handles(self):
        self.begin_handle = OccurrenceBeginHandle(self)
        self.end_handle = OccurrenceEndHandle(self)

    # FIXME: It is strange that this method must be called when adding an occurrence to the timeline. It seems that it ws neeeded during the creation of a new occurrence via the interface and got stuck in the creation of occurrences via the data file
    def end_creation(self):
        """Ends the creation of the occurrence"""
        self.update_selectable_flags()
        self.create_handles()

        # if begin_time is greater than end_time then swap times
        if self.begin_time > self.end_time:
            self.begin_time, self.end_time = self.end_time, self.begin_time
            self.update_rect()

        # Add this occurrence to the occurrence list of the timeline
        self.timeline.add_occurrence(self)

        self.update()

    def show_handles(self):
        if self.begin_handle:
            self.begin_handle.setVisible(True)
        if self.end_handle:
            self.end_handle.setVisible(True)

    def hide_handles(self):
        if self.begin_handle:
            self.begin_handle.setVisible(False)
        if self.end_handle:
            self.end_handle.setVisible(False)

    def get_bounds(self):
        _, lower_bound, upper_bound, occurrence = Occurrence.can_be_initiated(
            list(filter(lambda x: x != self, self.timeline.occurrences)),
            self.begin_time,
        )
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def get_time_from_bounding_interval(self, time) -> int:
        if self.lower_bound and time < self.lower_bound:
            time = self.lower_bound
        elif self.upper_bound and time > self.upper_bound:
            time = self.upper_bound
            self.time_pane.video.media_player.pause()
        return time


class OccurrenceHandle(QGraphicsRectItem):

    def __init__(self, occurrence: Occurrence, time: int, x: float):
        super().__init__(occurrence)
        self.occurrence = occurrence
        self.time = time

        self.pen: QPen = QPen(self.occurrence.pen_color)
        self.pen.setWidth(OCCURRENCE_PEN_WIDTH_OFF_CURSOR)
        self.setPen(self.pen)
        self.setBrush(self.occurrence.brush_color)
        self.setVisible(False)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)
        self.setAcceptDrops(True)

        self.height = (
            OCCURRENCE_HANDLE_HEIGHT_FRACTION * occurrence.rect().height()
        )
        self.setRect(
            QRectF(
                -OCCURRENCE_HANDLE_WIDTH / 2,
                -self.height / 2,
                OCCURRENCE_HANDLE_WIDTH,
                self.height,
            )
        )

        self.setX(x)
        self.setY(occurrence.rect().height() / 2)

    @abstractmethod
    def change_time(self, new_time):
        self.time = new_time
        self.occurrence.time_pane.data_needs_save = True

    def focusInEvent(self, event):
        self.occurrence.setSelected(True)
        self.occurrence.time_pane.video.set_position(self.time)
        self.pen.setWidth(OCCURRENCE_PEN_WIDTH_ON_CURSOR)
        self.setPen(self.pen)
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        self.occurrence.setSelected(False)
        self.pen.setWidth(OCCURRENCE_PEN_WIDTH_OFF_CURSOR)
        self.setPen(self.pen)
        super().focusOutEvent(event)

    def mouseMoveEvent(self, event):
        if event.buttons() == Qt.MouseButton.LeftButton:
            self.setY(self.height / 2)

            # With the mouse, the coordinate X is changed, but we need to
            # change the time.
            time = int(
                event.scenePos().x()
                * self.occurrence.time_pane.duration
                / self.occurrence.time_pane.scene.width()
            )

            time = self.occurrence.get_time_from_bounding_interval(time)

            self.occurrence.time_pane.video.set_position(time)


class OccurrenceBeginHandle(OccurrenceHandle):

    def __init__(self, occurrence: Occurrence):
        super().__init__(occurrence, occurrence.begin_time, 0)

    def change_time(self, time):
        t = time - int(self.occurrence.mfps / 2)
        super().change_time(t)
        self.occurrence.update_begin_time(t)


class OccurrenceEndHandle(OccurrenceHandle):
    def __init__(self, occurrence: Occurrence):
        super().__init__(
            occurrence, occurrence.end_time, occurrence.rect().width()
        )

    def change_time(self, time):
        t = time + int(self.occurrence.mfps / 2)
        super().change_time(t)
        self.occurrence.update_end_time(t)


class OccurrenceComment(QDialog):
    def __init__(self, text="", widget=None):
        super().__init__(widget)
        self.setWindowTitle("Occurrence comment")

        layout = QFormLayout(self)
        self.input = TextEdit(self, text)
        layout.addRow(self.input)

        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Cancel,
            self,
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    def get_text(self):
        return self.input.toPlainText()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import os
import re
import tempfile
import zipfile
import yaml

from PySide6.QtCore import QUrl
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QLabel,
    QMessageBox,
    QPushButton,
    QStyle,
    QVBoxLayout,
)
from PySide6.QtMultimedia import QMediaFormat
from .coders import Coders
from .config import Config
from .constants import (
    CONFIG_FILENAME,
    CSV_DELIMITER,
    METADATA_FILENAME,
)
from .data import Data
from .dialog import DialogCode
from .exceptions import LoadProjectError
from .format import FORMAT, format_ok


class Files:

    def __init__(self, window):

        self.window = window
        self.data_to_load = None
        self.csv_delimiter = CSV_DELIMITER
        self.config_file_name = CONFIG_FILENAME
        self.metadata_file_name = METADATA_FILENAME
        self.temp_dir = None
        self.file_format = None
        self.project_file_path = None
        self.coders = None

        # Search for supported video file formats
        self.video_file_extensions = []
        for f in QMediaFormat().supportedFileFormats(QMediaFormat.Decode):
            mime_type = QMediaFormat(f).mimeType()
            name = mime_type.name()
            if re.search("^video/", name):
                self.video_file_extensions.extend(mime_type.suffixes())
        extensions = " ".join(["*." + x for x in self.video_file_extensions])
        self.file_name_filters = [
            f"Video Files ({extensions})",
            "All Files (*.*)",
        ]

        self.project_file_filters = [
            f"Zip files ({' '.join(['*.zip'])})",
            "All Files (*.*)",
        ]

    def open_video(self, widget=None):
        """Open a video file in a MediaPlayer"""
        dialog_txt = "Open Video File"
        file_dialog = QFileDialog(widget) if widget else QFileDialog()
        file_dialog.setWindowTitle(dialog_txt)
        file_dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        file_dialog.setNameFilters(self.file_name_filters)
        file_dialog.exec()
        if file_dialog.result() == DialogCode.Accepted:
            # Load only the first of the selected file
            try:
                filename = file_dialog.selectedFiles()[0]
                self.load_video_file(filename)
                self.load_config_file()
            except Exception as e:
                QMessageBox.critical(self.window, "Error", f"{e}")

    def open_project(self, widget=None):
        """Open a project file"""
        dialog_txt = "Open Project File"
        project_dialog = QFileDialog(widget) if widget else QFileDialog()
        project_dialog.setWindowTitle(dialog_txt)
        project_dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        project_dialog.setNameFilters(self.project_file_filters)
        project_dialog.exec()
        if project_dialog.result() == DialogCode.Accepted:
            try:
                filename = project_dialog.selectedFiles()[0]
                self.load_project_file(filename)
            except Exception as e:
                QMessageBox.critical(self.window, "Error", f"{e}")

    def load_file(self, filename):
        try:
            if os.path.splitext(filename)[1] in [
                "." + ext for ext in self.video_file_extensions
            ]:
                self.load_video_file(filename)
                self.load_config_file()
            elif os.path.splitext(filename)[1] == ".zip":
                self.load_project_file(filename)
        except Exception as e:
            QMessageBox.critical(self.window, "Error", f"{e}")

    def load_video_file(self, filename):
        """Load video file"""
        if not os.path.exists(filename) or not os.path.isfile(filename):
            raise FileNotFoundError(
                f"FileNotFoundError : {filename} doesn't exist"
            )

        # getOpenFileName returns a tuple, so use only the actual file name
        self.media = QUrl.fromLocalFile(filename)
        self.window.video.media = self.media

        # Enable the buttons
        self.window.video.play_button.setEnabled(True)
        self.window.video.stop_button.setEnabled(True)

        # Put the media in the media player
        self.window.video.media_player.setSource(self.media)

        # Set the title of the track as window title
        self.window.setWindowTitle(os.path.basename(filename))

        # Show first frame
        self.window.video.media_player.play()
        self.window.video.media_player.pause()

        # Clear the time_pane
        if bool(self.media):
            self.window.time_pane.clear()
        self.window.time_pane.load_common()

    def load_project_file(self, filename):
        """Load project file"""
        if not os.path.exists(filename) or not os.path.isfile(filename):
            raise FileNotFoundError(f"FileNotFoundError : {filename} not found")

        # filename is a zip file
        self.project_file_path = filename

        with zipfile.ZipFile(filename, "r") as zip_file:
            # Create temp dir
            temp_dir = tempfile.TemporaryDirectory()

            # Load data from metadata.yml
            with zip_file.open(self.metadata_file_name) as f:
                data = yaml.load(f, Loader=yaml.FullLoader)
                self.file_format = data["format"]
                if not format_ok(self.file_format, self.window):
                    raise LoadProjectError("Format problem")
                files = zip_file.namelist()

                # Search for video in temp dir
                video_file = data["video"]
                if video_file not in files:
                    raise LoadProjectError("Failed to load video file")

                # Search for configuration file in temp dir
                if self.config_file_name not in files:
                    raise LoadProjectError("Failed to load config file")

                # Search for csv file in temp dir
                data_file = os.path.splitext(video_file)[0] + ".csv"
                if data_file not in files:
                    raise LoadProjectError("Failed to load data file")

                # Extract all files in temp dir
                zip_file.extractall(temp_dir.name)

                # Load video file from temp dir
                self.load_video_file(os.path.join(temp_dir.name, video_file))

                # Load config file from in temp dir
                self.load_config_file(
                    os.path.join(temp_dir.name, self.config_file_name)
                )

                # Load csv data file from in temp dir
                self.data_to_load = os.path.join(temp_dir.name, data_file)

            self.temp_dir = temp_dir

        self.window.time_pane.data_needs_save = False

    def load_config_file(self, filename=None):
        """load presets from configuration file"""
        # Read the YAML file
        config = Config() if filename is None else Config(filename)

        # Update the config file, eventually
        if self.file_format:
            config.update_format(self.file_format, FORMAT)

        self.window.time_pane.timelines_from_config(config)

        if "csv-delimiter" in config:
            self.csv_delimiter = config["csv-delimiter"]

        if "coders" in config:
            self.coders = Coders(config["coders"], self.window)

    def load_data_file(self, filename=None):
        """Load data file"""
        if os.path.isfile(filename):
            data = Data(filename, self.file_format, FORMAT)
            self.window.time_pane.add_timelines(data)
        else:
            QMessageBox.critical(
                self.window,
                "Error",
                "The file you tried to load does not exist.",
            )

        self.data_to_load = None

    def save_project(self) -> bool:
        """Save project file"""
        temp_dir = tempfile.TemporaryDirectory()

        # Construct the default file name from the QUrl of the video file
        target_directory = self.media.path()
        target_file_name = os.path.splitext(
            os.path.basename(self.media.path())
        )[0]
        data_file_name = target_file_name + ".csv"
        if self.project_file_path:
            target_directory = self.project_file_path
            target_file_name = os.path.splitext(
                os.path.basename(self.project_file_path)
            )[0]

        target_directory = (
            os.path.dirname(target_directory) + "/" + target_file_name + ".zip"
        )

        # 1. Create config file from information of time_pane in
        # tmp directory
        if self.coders:
            if not self.coders.current:
                self.coders.set_current()
        else:
            self.coders = Coders({}, self.window)
            self.coders.set_current()
        self.coders.current.set_date_now()

        config_file_path = os.path.join(temp_dir.name, self.config_file_name)
        self.export_config_file(config_file_path)

        # 2. Create CSV file "data.csv" from information of time_pane in
        # tmp directory
        data_file_path = os.path.join(temp_dir.name, data_file_name)
        self.export_data_file(data_file_path)

        metadata_file_path = os.path.join(
            temp_dir.name, self.metadata_file_name
        )
        with open(metadata_file_path, "w") as f:
            yaml.safe_dump(
                {
                    "video": os.path.basename(self.media.path()),
                    "format": FORMAT,
                },
                f,
            )

        # Open FileDialog
        path, _ = QFileDialog.getSaveFileName(
            self.window,
            "Save project",
            target_directory,
            "Zip Files (*.zip);;All Files (*)",
        )
        if path:
            with zipfile.ZipFile(path, "w") as zip_file:
                zip_file.write(metadata_file_path, self.metadata_file_name)
                zip_file.write(config_file_path, self.config_file_name)
                zip_file.write(data_file_path, data_file_name)
                zip_file.write(
                    self.media.toLocalFile(),
                    os.path.basename(self.media.path()),
                )
            self.window.time_pane.data_needs_save = False
            return True
        return False

    # FIXME: Move this into data.py (?)
    def export_data_file(self, target_path=None) -> bool:
        """Export data in CSV file"""
        if not target_path:
            if not self.is_exportable():
                QMessageBox.warning(
                    self.window, "No Data", "There is no data to save."
                )
                return False

            # Construct the default file name from the QUrl of the video file
            default_target_path = (
                os.path.dirname(self.media.path())
                + "/"
                + os.path.splitext(os.path.basename(self.media.path()))[0]
                + ".csv"
            )

            target_path, _ = QFileDialog.getSaveFileName(
                self.window,
                "Save data (CSV file)",
                default_target_path,
                "CSV Files (*.csv);;All Files (*)",
            )

        if target_path:
            df = self.window.time_pane.timelines_to_dataframe()
            df.to_csv(
                target_path,
                sep=self.csv_delimiter,
                encoding="utf-8",
                index=False,
            )
            self.window.time_pane.data_needs_save = False
            return True
        return False

    def is_exportable(self) -> bool:
        """Return true if the media file is exportable"""
        return (
            self.window.video.media_player is not None
            and self.window.time_pane is not None
            and self.window.time_pane.timelines
            and self.window.time_pane.has_occurrences()
        )

    def export_config_file(self, target_path=None):
        config = Config(target_path)

        timelines = self.window.time_pane.timelines
        nb_timelines = len(timelines)

        config["timelines"] = {
            t.name: {
                "order": nb_timelines - i,
                "description": t.description,
                "events": {
                    name: {
                        "color": t.event_collection.get_event(
                            name
                        ).color.name(),
                        "description": t.event_collection.get_event(
                            name
                        ).description,
                    }
                    for name in t.event_collection
                },
            }
            for i, t in enumerate(timelines)
        }

        if self.coders:
            config["coders"] = self.coders.to_list()

        # Write data
        config.save()

    def no_video_loaded(self):
        dialog = OpenProjectDialog(self.window)
        dialog.exec()

    def temp_dir_cleanup(self):
        if self.temp_dir:
            self.temp_dir.cleanup()


class OpenProjectDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Open a project")

        layout = QVBoxLayout(self)

        self.open_video_btn = QPushButton(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon),
            "Open video",
            self,
        )
        self.open_project_btn = QPushButton(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon),
            "Open project",
            self,
        )

        buttons = QDialogButtonBox(self)
        buttons.addButton(
            self.open_video_btn, QDialogButtonBox.ButtonRole.AcceptRole
        )
        buttons.addButton(
            self.open_project_btn, QDialogButtonBox.ButtonRole.AcceptRole
        )
        buttons.addButton(QDialogButtonBox.StandardButton.Cancel)

        buttons.clicked.connect(self.perform_action)
        buttons.rejected.connect(self.reject)

        layout.addWidget(
            QLabel("Choose a video or a project file to start coding")
        )
        layout.addWidget(buttons)
        self.setLayout(layout)

    def perform_action(self, button):
        if button == self.open_video_btn:
            self.parent().files.open_video(self)
        elif button == self.open_project_btn:
            self.parent().files.open_project(self)
        self.accept()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
from math import inf

from PySide6.QtCore import (
    QPointF,
    QRectF,
    Qt,
)
from PySide6.QtGui import (
    QPainter,
    QColor,
)
from PySide6.QtMultimedia import QMediaPlayer
from PySide6.QtWidgets import (
    QAbstractSlider,
    QGraphicsRectItem,
    QGraphicsScene,
    QGraphicsView,
    QMessageBox,
    QScrollBar,
    QVBoxLayout,
    QStyle,
    QWidget,
)
from PySide6.QtGui import QFontMetrics

from .constants import (
    CSV_HEADERS,
    EVENT_DEFAULT_COLOR,
    MINIMUM_ZOOM_FACTOR,
    OCCURRENCE_PEN_WIDTH_OFF_CURSOR,
    OCCURRENCE_PEN_WIDTH_ON_CURSOR,
    TIMELINE_HEIGHT,
    TIME_SCALE_HEIGHT,
    ZOOM_STEP,
)
from .cursor import Cursor
from .dialog import (
    ConfirmMessageBox,
    DialogCode,
    TimelineDialog,
)
from .event import (
    ChooseEvent,
    Event,
    EventCollection,
)
from .occurrence import (
    Occurrence,
    OccurrenceHandle,
)
from .ticklocator import TickLocator
from .timeline import Timeline


class ZoomableGraphicsView(QGraphicsView):

    def __init__(self, scene: QGraphicsScene, parent=None):
        super().__init__(scene, parent)
        self.zoom_factor = 1.0
        self.zoom_step = ZOOM_STEP
        self.zoom_shift = None
        self.minimum_zoom_factor = MINIMUM_ZOOM_FACTOR

        self.create_ui()

    def create_ui(self):
        vertical_scrollbar = QScrollBar(Qt.Orientation.Vertical, self)
        vertical_scrollbar.valueChanged.connect(
            self.on_vertical_scroll_value_changed
        )
        self.setVerticalScrollBar(vertical_scrollbar)

        self.setViewportUpdateMode(
            QGraphicsView.ViewportUpdateMode.FullViewportUpdate
        )
        self.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.setFrameShape(QGraphicsView.Shape.NoFrame)
        self.setAlignment(
            Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft
        )
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

    def wheelEvent(self, event):
        if event.modifiers() == Qt.KeyboardModifier.ControlModifier:
            if not self.parent().video.media:
                return
            mouse_pos = self.mapToScene(event.position().toPoint()).x()
            if event.angleDelta().y() > 0:
                self.zoom_shift = mouse_pos * (1 - self.zoom_step)
                self.zoom_in()
            else:
                self.zoom_shift = mouse_pos * (1 - 1 / self.zoom_step)
                self.zoom_out()
            self.zoom_shift = None
        elif event.modifiers() == Qt.KeyboardModifier.ShiftModifier:
            if event.angleDelta().y() > 0:
                action = QAbstractSlider.SliderSingleStepAdd
            else:
                action = QAbstractSlider.SliderSingleStepSub
            self.horizontalScrollBar().triggerAction(action)
        else:
            super().wheelEvent(event)

    def on_vertical_scroll_value_changed(self, value):
        if self.parent().time_pane_scale:
            self.parent().time_pane_scale.setPos(0, value)

    def zoom_in(self):
        self.zoom_factor *= self.zoom_step
        self.update_scale()

    def zoom_out(self):
        if self.zoom_factor / self.zoom_step >= self.minimum_zoom_factor:
            self.zoom_factor /= self.zoom_step
            self.update_scale()

    def update_scale(self):
        # Update the size of the scene with zoom_factor
        self.scene().setSceneRect(
            0,
            0,
            self.width() * self.zoom_factor,
            self.scene().height(),
        )

        if self.zoom_shift:
            previous_anchor = self.transformationAnchor()
            self.setTransformationAnchor(QGraphicsView.NoAnchor)
            self.translate(self.zoom_shift, 0)
            self.setTransformationAnchor(previous_anchor)

    def mouseDoubleClickEvent(self, event):
        super().mouseDoubleClickEvent(event)

        # Get the position click from the scene
        map = self.mapToScene(self.scene().sceneRect().toRect())
        x = map.boundingRect().x()

        # Compute the time of the position click
        time = int(
            (x + event.scenePosition().x())
            * self.parent().duration
            / self.scene().width()
        )

        self.parent().video.set_position(time)

    def set_position(self, time):
        # During the creation of a new occurrence
        if self.parent().occurrence_in_creation:
            occurence = self.parent().occurrence_in_creation
            time = occurence.get_time_from_bounding_interval(time)

        # Cope with selected occurrence
        for i in self.parent().scene.selectedItems():
            if isinstance(i, Occurrence):
                time = i.get_time_from_bounding_interval(time)
                break

        # Set time of the video player
        self.parent().window.video.media_player.setPosition(int(time))

    def keyPressEvent(self, event):
        pass


class TimePane(QWidget):

    def __init__(self, window):
        """Initializes the timeline widget"""
        super().__init__(window)
        self._duration = 0
        self._time = 0

        self.selected_timeline = None
        self.occurrence_in_creation: Occurrence = None
        self.window = window
        self.video = window.video
        self.timelines: list[Timeline] = []
        self.cursor = None
        self.time_pane_scale = None
        self.data_needs_save = False
        self.scrollbar_width = self.window.style().pixelMetric(
            QStyle.PM_ScrollBarExtent
        )

        self.setMouseTracking(True)

        self.create_ui()

    def create_ui(self):
        self.scene = QGraphicsScene()
        self.scene.sceneRectChanged.connect(self.on_scene_changed)
        self.scene.selectionChanged.connect(self.on_selection_changed)
        self.view = ZoomableGraphicsView(self.scene, self)
        self.scene.setSceneRect(0, 0, self.view.width(), self.view.height())

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.view)
        self.setLayout(layout)

    def on_scene_changed(self, rect):
        # Update occurrences
        for timeline in self.timelines:
            timeline.update_rect_width(rect.width())
            for occurrence in timeline.occurrences:
                occurrence.update_rect()

        if self.occurrence_in_creation:
            self.occurrence_in_creation.update_rect()

        # Update time_pane_scale display
        if self.time_pane_scale:
            # Update cursor
            if self.duration:
                self.time_pane_scale.cursor.setX(
                    self.time * rect.width() / self.duration
                )
            self.time_pane_scale.update_rect()

    def on_selection_changed(self):
        selected_items = self.scene.selectedItems()
        selected = None
        if len(selected_items) == 1:
            selected = selected_items[0]
            if isinstance(selected, Timeline):
                self.selected_timeline = selected
        for s in self.timelines:
            s.select = s == selected

    def select_cycle_timeline(self, delta, checked=False):
        i, n = self.find_selected_timeline()
        self.timelines[i].select = False
        if delta > 0:
            if i == n - 1:
                i = -1
        else:
            if i == 0:
                i = n
        i += delta
        self.select_timeline(self.timelines[i])

    def find_selected_timeline(self):
        n = len(self.timelines)
        for i in range(n):
            if self.timelines[i].select:
                break
        return i, n

    def select_timeline(self, line):
        for tl in self.timelines:
            tl.select = False
        line.select = True
        self.selected_timeline = line
        line.update()

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, time):
        if time != self._time:
            self._time = time
            self.on_time_changed(time)

    def on_time_changed(self, new_time):
        # First, update the current occurrence, if it exists. If the cursor
        # time goes beyond the allowed bounds, bring it back and do not update
        # the other widgets.
        if self.occurrence_in_creation:
            if (
                self.occurrence_in_creation.lower_bound
                and new_time < self.occurrence_in_creation.lower_bound
            ):
                new_time = self.occurrence_in_creation.lower_bound
            elif (
                self.occurrence_in_creation.upper_bound
                and new_time > self.occurrence_in_creation.upper_bound
            ):
                new_time = self.occurrence_in_creation.upper_bound
                if (
                    self.video.media_player.playbackState()
                    == QMediaPlayer.PlaybackState.PlayingState
                ):
                    self.video.media_player.pause()
            begin_time = self.occurrence_in_creation.begin_time
            end_time = self.occurrence_in_creation.end_time
            mfps = self.video.mfps
            if begin_time < end_time:
                if new_time >= begin_time:
                    self.occurrence_in_creation.update_end_time(
                        new_time + int(mfps / 2)
                    )
                else:
                    self.occurrence_in_creation.update_begin_time(end_time)
                    self.occurrence_in_creation.update_end_time(
                        begin_time - mfps
                    )
            else:
                if new_time <= begin_time:
                    self.occurrence_in_creation.update_end_time(
                        new_time - int(mfps / 2)
                    )
                else:
                    self.occurrence_in_creation.update_begin_time(end_time)
                    self.occurrence_in_creation.update_end_time(
                        begin_time + mfps
                    )

        # Cope with selected occurrence
        for i in self.scene.selectedItems():
            if isinstance(i, Occurrence):
                new_time = i.get_time_from_bounding_interval(new_time)
                break

        # Update cursor position
        if self.time_pane_scale and self.time_pane_scale.cursor:
            self.time_pane_scale.cursor.setX(
                new_time * self.scene.width() / self.duration
            )

        if isinstance(self.scene.focusItem(), OccurrenceHandle):
            occurrence_handle: OccurrenceHandle = self.scene.focusItem()
            occurrence_handle.change_time(new_time)

        # Change appearance of occurrence under the cursor
        # (Brute force approach; this ought to be improved)
        if not self.occurrence_in_creation:
            for t in self.timelines:
                for a in t.occurrences:
                    a.penWidth = OCCURRENCE_PEN_WIDTH_OFF_CURSOR
            if self.selected_timeline:
                for a in self.selected_timeline.occurrences:
                    if a.begin_time <= new_time and a.end_time >= new_time:
                        a.penWidth = OCCURRENCE_PEN_WIDTH_ON_CURSOR
                        break

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, duration):
        if duration != self._duration:
            self._duration = duration
            self.load_common()
            self.update()

    def load_common(self):
        # Recreate timeline
        self.time_pane_scale = TimePaneScale(self)

    def clear(self):
        # Clear timelineScene
        self.scene.clear()
        self.timelines = []

    # FIXME: [CREATION] move into occurrence.py
    def handle_occurrence(self):
        """Handle the occurrence"""
        if self.occurrence_in_creation is None:
            # FIXME: This bugs when there is no occurrences in timeline!!!
            can_be_initiate, lower_bound, upper_bound, occurrence = (
                Occurrence.can_be_initiated(
                    self.selected_timeline.occurrences, self.time
                )
            )
            if can_be_initiate:
                mfps = self.video.mfps
                self.occurrence_in_creation = Occurrence(
                    self.selected_timeline,
                    self.time - int(mfps / 2),
                    self.time + int(mfps / 2),
                    lower_bound,
                    upper_bound,
                )
            self.window.menu.add_occurrence_action.setText("Finish occurrence")
            self.window.menu.abort_occurrence_creation_action.setEnabled(True)
            if occurrence:
                occurrence.setSelected(not occurrence.isSelected())
                occurrence.get_bounds()

        else:
            # Finish the current occurrence
            events_dialog = ChooseEvent(
                self.selected_timeline.event_collection, self
            )
            events_dialog.exec()
            if events_dialog.result() == QMessageBox.DialogCode.Accepted:
                event = events_dialog.get_chosen()
                self.occurrence_in_creation.set_event(event)
                self.update()
                self.occurrence_in_creation.end_creation()
                self.occurrence_in_creation = None
                self.on_time_changed(self.time)
            menu = self.window.menu
            menu.add_occurrence_action.setText("Start occurrence")
            menu.abort_occurrence_creation_action.setEnabled(False)
            self.data_needs_save = True
            self.update()

    def handle_timeline(self):
        dialog = TimelineDialog(self)
        dialog.exec()
        if dialog.result() == DialogCode.Accepted:
            name = dialog.get_text()
            if name == "":
                QMessageBox.warning(
                    self, "Warning", "Timeline name cannot be empty"
                )
                return
            self.add_timeline(Timeline(name, self))
            self.data_needs_save = True

    def resizeEvent(self, a0):
        if self.time_pane_scale:
            origin = self.view.mapToScene(0, 0).x()
            width_before = self.scene.width() / self.view.zoom_factor
            width_after = self.view.width()
            shift = origin * (1 - width_after / width_before)
            self.view.update_scale()
            previous_anchor = self.view.transformationAnchor()
            self.view.setTransformationAnchor(QGraphicsView.NoAnchor)
            self.view.translate(shift, 0)
            self.view.setTransformationAnchor(previous_anchor)
        else:
            self.scene.setSceneRect(
                0,
                0,
                self.view.width(),
                TIME_SCALE_HEIGHT + TIMELINE_HEIGHT,
            )

        self.update()

    def abort_occurrence_creation(self):
        if self.occurrence_in_creation is not None:
            confirm_box = ConfirmMessageBox("Abort creation of occurrence?")
            if confirm_box.result() == ConfirmMessageBox.DialogCode.Accepted:
                self.occurrence_in_creation.remove()
                self.occurrence_in_creation = None
                self.update()
                menu = self.window.menu
                menu.abort_occurrence_creation_action.setEnabled(False)

    def add_timeline(self, timeline):
        self.timelines.append(timeline)
        self.add_to_scene(timeline)

        # Compute the new height of the scene
        new_height = TIME_SCALE_HEIGHT + len(self.timelines) * TIMELINE_HEIGHT
        scene_rect = self.scene.sceneRect()
        scene_rect.setHeight(new_height)
        self.scene.setSceneRect(scene_rect)

        # Set maximum height of the widget
        self.setMaximumHeight(int(new_height) + self.scrollbar_width + 2)

        # Select the new timeline
        for i in self.timelines:
            i.select = False
        timeline.select = True

    def add_to_scene(self, timeline):
        """Add the timeline to the scene"""
        # Set Y of the timeline based on the timescale height and the timeline
        # lines heights present on the scene
        timeline.setPos(
            0,
            TIME_SCALE_HEIGHT + (len(self.timelines) - 1) * TIMELINE_HEIGHT,
        )

        # Set the right rect based on the scene width and the height constant
        timeline.setRect(
            0,
            0,
            self.scene.width(),
            TIMELINE_HEIGHT,
        )

        # Add line to the scene
        self.scene.addItem(timeline)

    def add_timelines(self, data):
        for _, row in data.data.iterrows():
            # Search for timeline
            timeline = self.get_timeline_by_name(row["timeline"])

            # If timeline from csv doesn't exist in TimePane,
            # escape row
            if not timeline:
                continue

            # Search for event
            event = timeline.event_collection.get_event(row["event"])

            # If event from csv doesn't exist in timeline,
            # then add it
            if not event:
                continue

            occurrence = Occurrence(
                timeline,
                int(row["begin"]),
                int(row["end"]),
            )

            occurrence.set_event(event)
            occurrence.end_creation()

    def get_timeline_by_name(self, name):
        """Get the timeline by name"""
        return next((x for x in self.timelines if x.name == name), None)

    def has_occurrences(self) -> bool:
        return any(len(line.occurrences) for line in self.timelines)

    def delete_occurrence(self):
        for i in self.scene.selectedItems():
            if isinstance(i, Occurrence):
                i.on_remove()
                break

    def timelines_from_config(self, config):
        if "timelines" in config:

            # Set all absent order fields with Inf
            for k, v in config["timelines"].items():
                if not v:
                    v = dict()
                    config["timelines"][k] = v
                if "order" not in v:
                    v["order"] = -inf
                if "description" not in v:
                    v["description"] = ""

            # Sort according to order first and alphabetically from
            # timeline name, otherwise
            for item in sorted(
                config["timelines"].items(),
                key=lambda x: (x[1]["order"], x[0]),
            ):
                # Get name and properties of the timeline
                name = item[0]
                properties = item[1]

                # Create timeline
                line = Timeline(name, self)

                # Check description
                description = properties["description"]
                if description == "":
                    description = "(no description)"
                line.title.setToolTip(description)
                line.description = description

                # Add the timeline to the TimePane
                self.add_timeline(line)

                # Loop over events of the timeline
                event_collection = EventCollection()
                if "events" in properties:
                    for k, v in properties["events"].items():
                        event = Event(k)
                        try:
                            event.color = QColor(v["color"])
                        except KeyError:
                            event.color = EVENT_DEFAULT_COLOR
                        try:
                            event.description = v["description"]
                        except KeyError:
                            event.description = ""
                        event_collection.add_event(event)
                    line.event_collection = event_collection

    def timelines_to_dataframe(self):
        df = pd.DataFrame(columns=CSV_HEADERS)
        for timeline in sorted(
            self.window.time_pane.timelines, key=lambda x: x.name
        ):
            for occurrence in timeline.occurrences:
                comment = occurrence.comment.replace('"', '\\"')
                row = [
                    timeline.name,
                    occurrence.event.name,
                    occurrence.begin_time,
                    occurrence.end_time,
                    comment,
                ]
                df.loc[len(df.index)] = row
        return df


class TimePaneScale(QGraphicsRectItem):

    def __init__(self, time_pane: TimePane):
        super().__init__()
        self.time_pane = time_pane
        self.time_pane.scene.addItem(self)
        self.cursor = Cursor(self)
        self.setRect(
            QRectF(0, 0, self.time_pane.scene.width(), TIME_SCALE_HEIGHT)
        )

    def paint(self, painter, option, widget=...):
        self.draw_rect(painter)

        if self.time_pane.duration != 0:
            self.draw_scale(painter)

    def update_rect(self):
        self.setRect(
            QRectF(0, 0, self.time_pane.scene.width(), TIME_SCALE_HEIGHT)
        )
        self.update()

    def draw_rect(self, painter):
        """Draw the background rectangle of the timeline scale"""
        painter.setPen(Qt.GlobalColor.black)
        painter.setBrush(Qt.GlobalColor.lightGray)
        self.setRect(
            QRectF(0, 0, self.time_pane.scene.width(), TIME_SCALE_HEIGHT)
        )
        painter.drawRect(self.rect())

    def draw_scale(self, painter):
        tl = TickLocator()
        dur = self.time_pane.duration
        wid = self.time_pane.scene.width()
        font = painter.font()
        fm = QFontMetrics(font)
        loc = tl.find_locations(0, dur / 1000, wid, fm)
        # Compute the height of the text
        font_height = fm.height()
        line_height = 5
        y = self.rect().height()

        for p in loc:

            i = 1000 * (p[0] * wid / dur)

            # Compute the position of the text
            text_width = fm.boundingRect(p[1]).width()
            text_position = QPointF(i - text_width / 2, font_height)

            # Draw the text
            painter.drawText(text_position, p[1])

            # Compute the position of the line
            painter.drawLine(QPointF(i, y), QPointF(i, y - line_height))

    def mousePressEvent(self, event):
        return

    def mouseReleaseEvent(self, event):
        return



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""
ViCodePy
~~~~~~~~

Video coder in Python for Experimental Psychology

Usage:
  vicodepy [--size=<WIDTHxHEIGTH> | --fullscreen] [<path>]
  vicodepy -h | --help
"""

import sys

from docopt import docopt
from .app import App

usage = __doc__.split("\n\n")[2]


def main():
    """Entry point for the application"""
    args = docopt(usage, argv=sys.argv[1:])
    app = App(
        path=args["<path>"],
        size=args["--size"],
        fullscreen=args["--fullscreen"],
    )
    sys.exit(app.exec())



# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 6.6.2
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore

qt_resource_data = b"\
\x00\x00\x02\x04\
\x89\
PNG\x0d\x0a\x1a\x0a\x00\x00\x00\x0dIHDR\x00\
\x00\x00:\x00\x00\x00\x09\x08\x06\x00\x00\x00cU?+\
\x00\x00\x00\x09pHYs\x00\x00-\xd0\x00\x00-\xd0\
\x01\xe7\x9cB{\x00\x00\x00\x19tEXtSof\
tware\x00www.inksca\
pe.org\x9b\xee<\x1a\x00\x00\x01\x91ID\
ATH\x89\xe5\x95AK\x15Q\x18\x86\x9f\xebU\x17\
\xee\x14\x83(I7!BP\x7f \x8c \xbd\xb5J\
\xfc\x01\xf6\x03\xf2\xbf\xa9\x08\xa9a\x7f\xa1\x22\x02sg\
\xdd\x8c\xa8\xae!z7.\x9e\x163\xa3\xe7\x8e\xdf\x19\
\x09w\xf5\xc2,\xe6=\xcfy9g\xce\xf7\x9di\xa9\
\xfc\x0f\x1a\x0e\xbc\x19`\x018\x03\xb6\x80o\x0d\xf3\x87\
\x80\x0ep\x17\xf8\x08\xec\x02M_n\xaa\xe4\x05\xb6\x81\
n\x03\xdb\x02\x9e\x00s\xc0\xa7\x92o\xca\xbe\x09<\xa3\
\xd8\xd36p00\xaa\xa6\xcf\xb2\xda\xf7BG\xea|\
\x8d\xa9\x9eQu\xd7Am\xa8\xed\x0c\xffT=I\xd8\
cu1\xc3\xb6\xd5\xcdZ\xf6ku$\xc3?R\x7f\
'l_]J\x99\x14\x1eV{^\xd6^&|5\
`UW\x02\xb6\xa5~\x0d\xd8n9V\xe7_d\xb2\
_f\xd6\xb2\x1f\xb0\xbf\xca=\xa12\x94\x1c\xee40\
\x1e\x94\xc4,0\x16\xf8\xf73%\xf4 \xf0n\x00\xb7\
\x02\xff60y\xcd\xec1\x8a\xd6\xa9k\x02\xb8S\xbd\
\xa4\x1b=\x00z\xc1\x84=\xa0\x1f\xf8\xef2\x8by\x1b\
x?\x80\xc3\xc0\xef\x02?\x03\xff\xfd_d\xf7\x81\xfd\
\xc0\xef\x01\x9f\xcf\xdfj%\xb0\xa4\x9e&\xc7\x7f\xa4>\
\xcc\x94\xcb\xa8E\xdf\xa4Z3\xdf\xa3\x1d\x8b\xbe\xact\
\xac.d\xd8\xb6\xba^\xcb\xde1\xdf\xa3\xf3\xe5Z+\
\x9d\xaa\xcfS\xa6\xe5\xe5\xdf\xcb4\x17\xb7\xee+\xe0{\
\xe6\xebBq3.R\x94\xf7\x07\xe0M\x03\x0bE\xa9\
V\xb7\xee\x16\xf1)\xa7\xd9\x8f\x81{\x14U\xb5\xc3\xd5\
\xb7n\x07\x18)\xb3\xbf\x0c\x84\x05\x1b\xfd'\xf5\x07\xe1\
#\x85k\x01\x84\xa6V\x00\x00\x00\x00IEND\xae\
B`\x82\
"

qt_resource_name = b"\
\x00\x08\
\x0f\xa5\xa3\xf9\
\x00v\
\x00i\x00c\x00o\x00d\x00e\x00p\x00y\
\x00\x06\
\x06\x8a\x9c\xb3\
\x00a\
\x00s\x00s\x00e\x00t\x00s\
\x00\x0a\
\x0b-\x87\xc7\
\x00h\
\x00a\x00n\x00d\x00l\x00e\x00.\x00p\x00n\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x16\x00\x02\x00\x00\x00\x01\x00\x00\x00\x03\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00(\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x90U\x9b\xbc_\
"


def qInitResources():
    QtCore.qRegisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


def qCleanupResources():
    QtCore.qUnregisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


qInitResources()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import os

from PySide6.QtWidgets import (
    QApplication,
    QMessageBox,
)
from .mainwindow import MainWindow
from .constants import (
    APP_WINDOW_WIDTH,
    APP_WINDOW_HEIGHT,
)


class App(QApplication):

    def __init__(self, path=None, size=None, fullscreen=None):
        super().__init__([])
        window = MainWindow()
        width = APP_WINDOW_WIDTH
        height = APP_WINDOW_HEIGHT
        self.size = size
        if self.size:
            dim = self.size.split("x")
            if len(dim) == 2:
                try:
                    width = int(dim[0])
                    height = int(dim[1])
                except ValueError:
                    self.wrong_size_message()
            else:
                self.wrong_size_message()
        if fullscreen:
            window.showFullScreen()
        else:
            window.resize(width, height)
        window.show()
        if path:
            window.files.load_file(os.path.abspath(path))
        else:
            window.files.no_video_loaded()

    def wrong_size_message(self):
        msgbox = QMessageBox()
        msgbox.setWindowTitle("ViCodePy: parse error of size argument")
        msgbox.setIcon(QMessageBox.Warning)
        msgbox.setText(
            f"The value given to the --size option ({self.size}) is not in "
            "the format 'WIDTHxHEIGHT', with WIDTH and HEIGHT being integer "
            "numbers. Using the default value "
            f"{self.DEFAULT_WIDTH}x{self.DEFAULT_HEIGHT}."
        )
        msgbox.setStandardButtons(QMessageBox.Ok)
        msgbox.exec()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.


class LoadProjectError(Exception):
    """Custom exception for errors occurring while loading a project."""

    def __init__(self, message="An error occurred while loading the project"):
        super().__init__(message)



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import pandas as pd


class Data:
    def __init__(self, filename, file_format, current_format):
        self.data = pd.read_csv(filename, sep=None, engine="python")
        self.update_format(file_format, current_format)

    def update_format(self, file_format, current_format):
        for fmt in range(file_format + 1, current_format + 1):
            if fmt == 2:
                pass
            if fmt == 3:
                self.data.rename(columns={"label": "event"}, inplace=True)
